import binascii
import json
import logging
import os.path
import ssl
import struct
import subprocess
from http.client import parse_headers
from urllib.parse import urlparse

import aiohttp
from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.hazmat.primitives.serialization import load_der_private_key
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives.serialization import NoEncryption
from cryptography.hazmat.primitives.serialization import PrivateFormat

from .config import backup
from .id6 import Id6

# Note to test:
# cd .../basicstation/examples/cups/
# python3 sim.py
# runs a CUPS server simulator

log = logging.getLogger('cups')


class CUPSClient():
    '''BasicStation Configuration and UPdate Server client'''

    def __init__(self, routerid, homedir):
        self.routerid = Id6(routerid)
        self.homedir = homedir
        self.uri = {'cups': None, 'tc': None}
        self.cred = {'cups': {}, 'tc': {}}

    def update_uri(self, service, uri):
        log.info(f"New {service.upper()} URI: {uri}")
        self.uri[service] = uri
        urifile = os.path.join(self.homedir,  service + ".uri")
        with open(urifile, "wb") as fd:
            fd.write(uri.encode())

    def update_cred(self, service, cred):
        log.info(f"New {service.upper()} credentials: {cred}")
        cred.write_files(service, self.cred[service], folder=self.homedir)

    async def roll_back(self, suffix):
        # roll back cups credentials from either -bak or -boot
        # file set onto current file set
        log.info(f"rolling back cups credentials from {suffix} file set")
        backup("cups", suffix, revert=True)

    async def update_info(self, cuuri, cutrust=None, cucrt=None, cukey=None,
                          tcuri=None, tctrust=None, tccrt=None, tckey=None,
                          capath=None, versions={}, info_max=1024,
                          auto_backup=True):
        self.cred['cups']['trust'] = cutrust
        self.cred['cups']['crt'] = cucrt
        self.cred['cups']['key'] = cukey
        self.cred['tc']['trust'] = tctrust
        self.cred['tc']['crt'] = tccrt
        self.cred['tc']['key'] = tckey

        self.uri['cups'] = cuuri
        self.uri['tc'] = tcuri

        try:
            cupscred = Credentials.from_files(cutrust, cucrt, cukey)
        except ValueError as tce:
            log.error("Could not load CUPS credentials : " + str(tce))
            cupscred = Credentials.from_files(None, None, None)
        try:
            tccred = Credentials.from_files(tctrust, tccrt, tckey)
        except ValueError as tce:
            log.error("Could not load LNS credentials : " + str(tce))
            tccred = Credentials.from_files(None, None, None)

        log.info("Update info to " + cuuri)
        log.debug("CUPS URI:{}, Credentials: {}; TC URI: {}, Credentials {}"
                  .format(cuuri, cupscred, tcuri, tccred))

        pkg_version = versions.get('package')
        log.debug("PKG VER : " + pkg_version)

        if len(pkg_version) > info_max:
            pkg_version = pkg_version.replace(".", "")
            if len(pkg_version) > info_max:
                pkg_version = pkg_version.replace("lorad", "ld")

        info = {
            'router': str(self.routerid),
            'cupsUri': self.uri['cups'],
            'tcUri': self.uri['tc'],
            'cupsCredCrc': cupscred.crc(),
            'tcCredCrc': tccred.crc(),
            'station': versions.get('station'),
            'model': versions.get('model'),
            'package': pkg_version,
            'keys': []        # TODO Not supported
        }

        # Return True if an update was received
        updated = {"cups": False, "tc": False, "update": False}

        uri = self.uri['cups'] + "/update-info"
        httpargs = connection_args(cutrust, cucrt, cukey, capath, 'headers')
        if {} == httpargs:
            return None

        handle_firewall(uri)

        try:
            async with aiohttp.ClientSession() as session:
                async with session.post(uri, json=info, **httpargs) as resp:
                    status = resp.status
                    data = await resp.read()
            _err_msg_ = None
        except aiohttp.ClientConnectorError:
            log.warning("CUPS POST EXCEPTION")
            return None
        try:
            Data = json.loads(data)
            for K in Data.keys():
                if K.lower() == "message":
                    _err_msg_ = Data[K]
                    break
        except ValueError:
            _err_msg_ = None

        if (status == 200) and (_err_msg_ is None):
            # Cups transaction was successfull, back up the
            # CUPS credentials before any potential update
            if auto_backup is True:
                backup("cups", "-bak")

            # Decode URIs
            for service in ["cups", "tc"]:
                uriLen = data[0]
                data = data[1:]
                if uriLen:
                    uri = data[:uriLen].decode('utf-8')
                    data = data[uriLen:]
                    self.update_uri(service, uri)
                    updated[service] = True

            # Decode credentials
            for service in ["cups", "tc"]:
                credLen, = struct.unpack_from("<H", data)
                data = data[2:]
                if credLen:
                    cred = Credentials.from_blob(data[:credLen])
                    data = data[credLen:]
                    self.update_cred(service, cred)
                    updated[service] = True

            # TODO: signLen (4B), keyCRC (4B), signature, updLen (4B), updData
        else:
            if _err_msg_ is not None:
                log.error(f"Update info error: {_err_msg_} , status {status}")
                if info_max == 1024:
                    # Try to limit the info size in case of AWS server ..
                    return await self.update_info(cuuri, cutrust, cucrt,
                                                  cukey, tcuri, tctrust,
                                                  tccrt, tckey, capath,
                                                  versions, info_max=32,
                                                  auto_backup=auto_backup)
            else:
                log.error(f"Update info error: HTTP response status: {status}")
            if 401 == status:  # can happen when key is corrupted
                return None
        return updated


class Credentials():
    '''Class to handle CUPS credentials encoding, decoding and verification'''

    def __init__(self):
        '''Store credentials in cryptography's format'''
        self.trust = None
        self.crt = None
        self.key = None

    def __repr__(self):
        return '{self.__class__.__name__}({self.trust}, {self.crt},'\
               ' {self.key}) ' \
               'CRC32 0x{crc:08X} ({crc})'.format(self=self, crc=self.crc())

    def __str__(self):
        return '0x{crc:08X} ({crc})'.format(crc=self.crc())

    @classmethod
    def from_blob(cls, blob):
        cred = cls()
        cred.parse_binary(blob)
        return cred

    @classmethod
    def from_files(cls, trust=None, crt=None, key=None):
        cred = cls()
        cred.load_files(trust, crt, key)
        return cred

    def parse_binary(self, blob):
        '''CUPS provides a credential blob containing a concatenation of trust/crt/key
        in DER format. Extract and store.'''
        try:
            self.trust = x509.load_der_x509_certificate(
                blob, backend=default_backend())
        except ValueError:
            log.error("Credentials parse error: invalid `trust`")
            return

        trust_len = len(self.trust.public_bytes(Encoding.DER))
        blob = blob[trust_len:]
        try:
            self.crt = x509.load_der_x509_certificate(
                blob, backend=default_backend())
        except ValueError:
            if blob[:4] == b'\x00'*4:
                self.crt = None
            else:
                log.error("Credentials parse error: invalid `crt`")
                return
        if self.crt:
            crt_len = len(self.crt.public_bytes(Encoding.DER))
            blob = blob[crt_len:]
        else:
            self.key = blob[4:]
            return

        try:
            self.key = load_der_private_key(
                blob, password=None, backend=default_backend())
        except ValueError:
            log.error("Credentials parse error: invalid `key`")
            return

    def load_files(self, trust=None, crt=None, key=None):
        self.trust = self.crt = self.key = None  # Default
        if trust:
            with open(trust, 'rb') as f:
                self.trust = x509.load_pem_x509_certificate(
                    f.read(), default_backend())

        if crt:
            with open(crt, 'rb') as f:
                self.crt = x509.load_pem_x509_certificate(
                    f.read(), default_backend())

        if crt and key:
            with open(key, 'rb') as f:
                self.key = load_pem_private_key(
                    f.read(), None, default_backend())
        elif key:
            with open(key, 'rb') as f:
                self.key = f.read()

    def write_files(self, service, credpath, folder='.'):
        filename = os.path.join(folder, service)
        if self.trust:
            f = filename + ".trust"
            with open(f, "wb") as fd:
                fd.write(self.trust.public_bytes(Encoding.PEM))
            log.debug(f + " written")
            credpath['trust'] = f

        if self.crt:
            f = filename + ".crt"
            with open(f, "wb") as fd:
                fd.write(self.crt.public_bytes(Encoding.PEM))
            log.debug(f + " written")
            credpath['crt'] = f

        if self.crt and self.key:
            f = filename + ".key"
            with open(f, "wb") as fd:
                fd.write(self.key.private_bytes(
                    Encoding.PEM,
                    PrivateFormat.TraditionalOpenSSL,
                    NoEncryption()))
            log.debug(f + " written  (private key)")
            credpath['key'] = f
        elif self.key:
            f = filename + ".key"
            with open(f, "wb") as fd:
                fd.write(self.key)
            log.debug(f + " written (token)")
            credpath['key'] = f

    def crc(self):
        data = b''
        if self.trust:
            data += self.trust.public_bytes(Encoding.DER)
        else:
            data += b'\x00'*4
        if self.crt:
            data += self.crt.public_bytes(Encoding.DER)
        else:
            data += b'\x00'*4

        if getattr(self.key, "private_bytes", None):
            data += self.key.private_bytes(Encoding.DER,
                                           PrivateFormat.TraditionalOpenSSL,
                                           NoEncryption())
        elif self.key:
            data += self.key
        else:
            data += b'\x00'*4

        return binascii.crc32(data)


def connection_args(trust=None, crt=None, key=None,
                    capath=None, header_arg='extra_headers'):
    '''This helper function create the appropriate argument list for
    websocket and http connections.
    Use the returned dict as connect(..., **args) to pass it as kwargs
    **args allow to pass 'ssl' kwargs only if necessary. ssl=None has a meaning
    NOTE: websockets uses 'extra_headers', aiohttp 'headers' for header_arg'''
    # TLS and Authentication
    args = {}
    try:
        if trust:  # Server TLS auth
            sslcontext = ssl.create_default_context(
                cafile=trust, capath=capath)
            args.update({'ssl': sslcontext})
        if crt:  # client TLS auth
            if not trust:  # No trust file => fall back to default
                sslcontext = ssl.create_default_context()
            sslcontext.load_cert_chain(crt, key)
            args.update({'ssl': sslcontext})
        elif key:  # key without crt => key file contains HTTP headers
            with open(key, 'rb') as fd:
                headers = dict(parse_headers(fd))
            args.update({header_arg: headers})
    except ssl.SSLError as e:
        log.error("SSL Error " + str(e))
        return {}

    return args


def handle_firewall(uri):
    # Verify URI (will raise Value Error if something is wrong)
    uriInfo = urlparse(uri)

    # Open Firewall if needed
    if uriInfo.port:
        outrule = "OUTPUT -p tcp --dport " + str(uriInfo.port) + " -j ACCEPT"
        inrule = "INPUT -p tcp --sport " + str(uriInfo.port) + " -j ACCEPT"

        # Try delete before adding
        # (Yes, Shell call is Ugly but python3-netfilter is not installed ...
        #  So it's still smarter than Shell parsing in init script)
        subprocess.call("iptables -D " + outrule, shell=True)
        subprocess.call("iptables -D " + inrule, shell=True)

        subprocess.call("iptables -I " + outrule, shell=True)
        subprocess.call("iptables -I " + inrule, shell=True)
