import asyncio
import logging
import os
import sqlite3
import sys
import time
from sqlite3 import Error

if sys.version_info[1] < 7:
    asyncio.create_task = asyncio.ensure_future

log = logging.getLogger('database')

DBVERSION = 1
PACKAGE_VERSION = "1.0.0"


class DatabaseClient():
    '''Basic Station Database client (sqlite)'''

    def __init__(self, depth, base):
        log.info("DatabaseClient Init")
        self.db_conn = None
        self.database = base
        self.depth = depth
        # allow an extra 10% size (or 5 messages if configured
        # database depth is really small)
        self.margin = max(5, int(depth / 10))

    async def create_connection(self):
        os.makedirs(os.path.dirname(self.database), exist_ok=True)
        log.info("creating connection to " + str(self.database))
        connection = None
        try:
            connection = sqlite3.connect(self.database)
            log.info("Connection to SQLite DB successful")
        except Error as e:
            log.error(f"The error '{e}' occurred")
        self.db_conn = connection

        await self.check_version()

    async def create_database(self):
        res1 = await self.execute_query("CREATE TABLE IF NOT EXISTS versions\
         ( database INTEGER NOT NULL, binary TEXT NOT NULL UNIQUE);")
        res2 = await self.execute_query("CREATE TABLE IF NOT EXISTS messages\
         ( id INTEGER PRIMARY KEY AUTOINCREMENT, recv_time INTEGER NOT NULL \
         DEFAULT 0, loramsg BLOB NOT NULL, frequency INTEGER, spreading_factor\
          INTEGER, bandwidth INTEGER, receive_context INTEGER, xtime INTEGER, \
         rssi REAL, snr REAL);")
        return res1 + res2

    async def execute_query(self, query, data=None):
        log.debug("QUERY : [ "+query+" ]")
        cursor = self.db_conn.cursor()
        try:
            if data is None:
                cursor.execute(query)
            else:
                cursor.execute(query, data)
            self.db_conn.commit()
            return 0
        except Error as e:
            log.error(f"The error '{e}' occurred")
            return 1

    async def execute_read_query(self, query):
        log.debug("READ QUERY : [ "+query+" ]")
        cursor = self.db_conn.cursor()
        result = None
        try:
            cursor.execute(query)
            result = cursor.fetchall()
        except Error as e:
            log.error(f"The error '{e}' occurred")
        return result

    async def update(self, V):
        return True

    async def size(self):
        res = await self.execute_read_query("SELECT count(*) FROM messages")
        log.debug("SIZE : " + str(res))
        return res[0][0]

    async def clear(self):
        return await self.execute_query("DELETE FROM messages")

    async def erase(self, msg_id):
        ERASE_QUERY = "DELETE FROM messages WHERE id = " + str(msg_id)
        return await self.execute_query(ERASE_QUERY)

    async def check_version(self):
        res = await self.execute_read_query("SELECT database, binary FROM \
        versions ORDER BY rowid DESC LIMIT 1")

        if (res is None) or (len(res) == 0):
            log.info("Creating " + self.database + " database")
            await self.create_database()
        else:
            [dbversion, version] = res[0]
            if dbversion > DBVERSION:
                log.warning("Database version not supported: " +
                            str(dbversion) + " > " + str(DBVERSION))
                return False
            if (dbversion < DBVERSION) and (not self.update(dbversion)):
                log.error("Database version update failed")
                return False
        return await self.execute_query("INSERT INTO versions(database, \
        binary) VALUES('"+str(DBVERSION)+"','"+PACKAGE_VERSION+"')")

    async def free_space_up(self, db_size):
        if db_size > self.depth:
            # get rid of the oldest 9% at most (D*10% / (D*10% + D))
            old_msgs_nb = db_size - self.depth
            log.info("erasing " + str(old_msgs_nb) + " oldest messages ..")

            NTH_OLDEST_QUERY = "SELECT recv_time FROM messages ORDER BY recv_time \
                                ASC LIMIT 1 OFFSET " + str(old_msgs_nb)
            min_tmsp = (await self.execute_read_query(NTH_OLDEST_QUERY))[0][0]
            log.debug("keeping messages from : " + str(min_tmsp))

            CLEAN_UP_QUERY = "DELETE FROM messages WHERE recv_time < " + \
                             str(min_tmsp)
            return await self.execute_query(CLEAN_UP_QUERY)

    async def insert(self, loramsg, freq, sf, bw, rctx,
                     xtime, rssi, snr, rcvtime=None):

        if rcvtime is None:
            # store current buffering time, in GPS time ms
            rcvtime = 1000000 * (int(time.time()) - 315964819)

        db_size = await self.size()
        if db_size > (self.depth + self.margin):
            log.warning("FREEING UP SOME SPACE FIRST ..")
            await self.free_space_up(db_size)

        log.debug("++++ INSERTING " + str(loramsg) + " INTO DB")
        QUERY = "INSERT INTO messages(recv_time, loramsg, frequency, \
        spreading_factor, bandwidth, receive_context, xtime, rssi, snr) \
        VALUES("+str(rcvtime)+",?,"+str(freq)+",'"+str(sf)+"','"+str(bw)+"',"\
            + str(rctx)+","+str(xtime)+","+str(rssi)+","+str(snr)+")"

        return await self.execute_query(QUERY, (loramsg,))

    async def fetch(self, nb):
        # DO WE NEED SOME SORT OF MUTEX BETWEEN FETCH AND INSERT ??

        db_size = await self.size()
        if db_size > (self.depth + self.margin):
            log.debug("FREEING UP EXTRA MESSAGES")
            await self.free_space_up(db_size)

        Q = "SELECT id, recv_time, loramsg, frequency, spreading_factor, \
        bandwidth, receive_context, xtime, rssi, snr FROM messages \
        ORDER BY recv_time ASC LIMIT " + str(nb)

        MSGS = await self.execute_read_query(Q)

        retList = list()
        for M in MSGS:
            (ID, rcvtmsp, loramsg, freq, sf, bw, rctx, xtime, rssi, snr) = M

            log.debug("M : " + str(M))

            retList.append([ID, [loramsg, freq, sf, bw, rctx,
                           xtime, rssi, snr, rcvtmsp]])

        return retList
